@@ -23,6 +23,7 @@ with this program.  If not, see <http://www.gnu.org/licenses/>.
 #include "config.h"
 #endif
 
+#include <math.h>
 #include <stdlib.h>
 #include <glib.h>
 
@@ -105,6 +106,7 @@ struct _UbuntuGeoipProviderPrivate
 #define UBUNTU_GEOIP_PROVIDER_GET_PRIVATE(o) \
 (G_TYPE_INSTANCE_GET_PRIVATE ((o), UBUNTU_GEOIP_PROVIDER_TYPE, UbuntuGeoipProviderPrivate))
 
+static gchar * test_data = NULL;
 static GMainLoop * mainloop = NULL;
 static const gchar * field_names[MARKUP_FIELD_CNT] = {
 	"CountryCode",
@@ -155,6 +157,10 @@ static void markup_text                      (GMarkupParseContext * context,
 static void url_changed                      (GSettings * settings,
                                               char * key,
                                               gpointer user_data);
+static gboolean parse_xml_data               (UbuntuGeoipProvider * provider,
+                                              const gchar * xml,
+                                              const gint length);
+static gboolean use_test_data                (gpointer user_data);
 
 static GMarkupParser parser_struct = {
 	start_element: markup_start,
@@ -244,12 +250,17 @@ ubuntu_geoip_provider_init (UbuntuGeoipProvider *self)
 	}
 
 	/* Allocate */
-	self->priv->soup = soup_session_async_new();
-	self->priv->client = nm_client_new();
-	g_signal_connect(G_OBJECT(self->priv->client), "notify::" NM_CLIENT_STATE, G_CALLBACK(nm_state_change), self);
+	if (test_data == NULL) {
+		self->priv->soup = soup_session_async_new();
+		self->priv->client = nm_client_new();
+		g_signal_connect(G_OBJECT(self->priv->client), "notify::" NM_CLIENT_STATE, G_CALLBACK(nm_state_change), self);
 
-	/* Check things out */
-	nm_state_change(self->priv->client, NULL, self);
+		/* Check things out */
+		nm_state_change(self->priv->client, NULL, self);
+	} else {
+		self->priv->network_state = NM_STATE_CONNECTED_GLOBAL;
+		g_idle_add(use_test_data, self);
+	}
 
 	gc_provider_set_details (GC_PROVIDER (self),
 	                         "org.freedesktop.Geoclue.Providers.UbuntuGeoIP",
@@ -336,6 +347,36 @@ ubuntu_geoip_provider_finalize (GObject *object)
 	return;
 }
 
+/* Function to load the test data and pass it into the parser */
+static gboolean
+use_test_data (gpointer user_data)
+{
+	g_return_val_if_fail(IS_UBUNTU_GEOIP_PROVIDER(user_data), FALSE);
+	g_return_val_if_fail(test_data != NULL, FALSE);
+
+	gchar * data = NULL;
+	gsize length = 0;
+
+	if (g_file_test(test_data, G_FILE_TEST_EXISTS)) {
+		GError * error = NULL;
+
+		g_file_get_contents(test_data, &data, &length, &error);
+
+		if (error != NULL) {
+			g_warning("Unable to read test data '%s': %s", test_data, error->message);
+			g_error_free(error);
+		}
+	}
+
+	if (data != NULL) {
+		parse_xml_data(UBUNTU_GEOIP_PROVIDER(user_data), data, length);
+
+		g_free(data);
+	}
+
+	return FALSE;
+}
+
 /* Something could change our URL settings, if so we need to respond
    and start getting the new data */
 static void
@@ -386,6 +427,33 @@ get_status (GcIfaceGeoclue *geoclue, GeoclueStatus * status, GError ** error)
 	return TRUE;
 }
 
+/* Figure out how accurate we are based on the information we have */
+GeoclueAccuracy *
+position_accuracy_build (UbuntuGeoipProvider * provider)
+{
+	GeoclueAccuracy * accuracy = geoclue_accuracy_new (GEOCLUE_ACCURACY_LEVEL_NONE, 0.0, 0.0);
+
+	/* We're basically making fixed point values at the 10K location */
+	gint fixed_alat = abs(provider->priv->latitude * 10000.0);
+	gint fixed_alon = abs(provider->priv->longitude * 10000.0);
+
+	/* Basically what we're doing here is checking the rounding.  If we're
+	   rounding to an even value... it's roughly a country. */
+	if ((fixed_alat % 100000) > 0 || (fixed_alon % 100000) > 0) { /* 1.0 accuracy */
+		geoclue_accuracy_set_details(accuracy, GEOCLUE_ACCURACY_LEVEL_COUNTRY, 0.0, 0.0);
+
+		if ((fixed_alat % 10000) > 0 || (fixed_alon % 10000) > 0) { /* 0.1 accuracy */
+			geoclue_accuracy_set_details(accuracy, GEOCLUE_ACCURACY_LEVEL_REGION, 0.0, 0.0);
+
+			if ((fixed_alat % 1000) > 0 || (fixed_alon % 1000) > 0) { /* 0.01 accuracy */
+				geoclue_accuracy_set_details(accuracy, GEOCLUE_ACCURACY_LEVEL_LOCALITY, 0.0, 0.0);
+			}
+		}
+	}
+
+	return accuracy;
+}
+
 /* Get the current position as we see it. */
 static gboolean
 get_position (GcIfacePosition *gc, GeocluePositionFields * fields, int * timestamp, double *latitude, double *longitude, double *altitude, GeoclueAccuracy **accuracy, GError **error)
@@ -402,7 +470,7 @@ get_position (GcIfacePosition *gc, GeocluePositionFields * fields, int * timesta
 		*fields = GEOCLUE_POSITION_FIELDS_LATITUDE | GEOCLUE_POSITION_FIELDS_LONGITUDE;
 		*latitude = provider->priv->latitude;
 		*longitude = provider->priv->longitude;
-		*accuracy = geoclue_accuracy_new (GEOCLUE_ACCURACY_LEVEL_LOCALITY, 0.0, 0.0);
+		*accuracy = position_accuracy_build(provider);
 	}
 
 	return TRUE;
@@ -416,6 +484,27 @@ position_iface_init (GcIfacePositionClass *iface)
 	return;
 }
 
+/* Figure out how accurate we are based on the information we have */
+GeoclueAccuracy *
+address_accuracy_build (UbuntuGeoipProvider * provider)
+{
+	GeoclueAccuracy * accuracy = geoclue_accuracy_new (GEOCLUE_ACCURACY_LEVEL_NONE, 0.0, 0.0);
+
+	if (provider->priv->country_name != NULL || provider->priv->country_code != NULL) {
+		geoclue_accuracy_set_details(accuracy, GEOCLUE_ACCURACY_LEVEL_COUNTRY, 0.0, 0.0);
+	}
+
+	if (provider->priv->region_name != NULL) {
+		geoclue_accuracy_set_details(accuracy, GEOCLUE_ACCURACY_LEVEL_REGION, 0.0, 0.0);
+	}
+
+	if (provider->priv->city != NULL || provider->priv->postal_code != NULL) {
+		geoclue_accuracy_set_details(accuracy, GEOCLUE_ACCURACY_LEVEL_LOCALITY, 0.0, 0.0);
+	}
+
+	return accuracy;
+}
+
 /* Put the details into the hash table correctly */
 static void
 address_details_build (UbuntuGeoipProvider * provider, GHashTable * details)
@@ -462,7 +551,7 @@ get_address (GcIfaceAddress *gc, int *timestamp, GHashTable **address, GeoclueAc
 	} else {
 		*address = geoclue_address_details_new();
 		address_details_build(provider, *address);
-		*accuracy = geoclue_accuracy_new (GEOCLUE_ACCURACY_LEVEL_LOCALITY, 0.0, 0.0);
+		*accuracy = address_accuracy_build(provider);
 	}
 
 	return TRUE;
@@ -476,6 +565,88 @@ address_iface_init (GcIfaceAddressClass *iface)
 	return;
 }
 
+/* Take the XML data we got from the internet and turn it into
+   something useful */
+static gboolean
+parse_xml_data (UbuntuGeoipProvider * provider, const gchar * xml, const gint length)
+{
+	g_return_val_if_fail(xml != NULL, FALSE);
+	g_return_val_if_fail(length > 10, FALSE);
+
+	provider->priv->markup_data.position_changed = FALSE;
+	provider->priv->markup_data.address_changed = FALSE;
+	provider->priv->markup_data.valid_header = FALSE;
+
+	/* Parse through the data using the markup functions already setup in
+	   the context */
+	if (g_markup_parse_context_parse(provider->priv->context, xml, length, NULL) == FALSE) {
+		g_warning("Unable to parse the data from the geoip provider");
+
+		/* Once we've broken a context, clear it for next time */
+		g_clear_object(&provider->priv->context);
+		provider->priv->context = g_markup_parse_context_new(&parser_struct, 0, &provider->priv->markup_data, NULL);
+
+		return FALSE;
+	}
+
+	/* The parser won't check the format, this field ensures we got
+	   a format that we expect */
+	if (!provider->priv->markup_data.valid_header) {
+		g_warning("We didn't get valid XML, that's not cool.");
+		return FALSE;
+	}
+
+	/* We only need to emit a state change if we're switching from
+	   not having data to having it */
+	gboolean emit = FALSE;
+	if (provider->priv->gotdata == FALSE) {
+		emit = TRUE;
+	}
+	provider->priv->gotdata = TRUE;
+
+	/* If the position changed, update that */
+	if (provider->priv->markup_data.position_changed) {
+		g_debug("New position:");
+		g_debug("Latitude:  %f", provider->priv->latitude);
+		g_debug("Longitude: %f", provider->priv->longitude);
+
+		GeoclueAccuracy * accuracy = position_accuracy_build(provider);
+		gc_iface_position_emit_position_changed(GC_IFACE_POSITION(provider), 
+		                                        time(NULL),
+		                                        GEOCLUE_POSITION_FIELDS_LATITUDE | GEOCLUE_POSITION_FIELDS_LONGITUDE,
+		                                        provider->priv->latitude,
+		                                        provider->priv->longitude,
+		                                        0, /* Altitude */
+		                                        accuracy);
+
+		geoclue_accuracy_free(accuracy);
+	}
+
+	/* If the address changed, update that */
+	if (provider->priv->markup_data.address_changed) {
+		GeoclueAccuracy * accuracy = address_accuracy_build(provider);
+		GHashTable * address = geoclue_address_details_new();
+
+		address_details_build(provider, address);
+
+		gc_iface_address_emit_address_changed(GC_IFACE_ADDRESS(provider), 
+		                                        time(NULL),
+		                                        address,
+		                                        accuracy);
+
+		g_hash_table_destroy(address);
+		geoclue_accuracy_free(accuracy);
+	}
+
+	/* Now emit our status */
+	if (emit) {
+		g_debug("Emitting Available");
+		gc_iface_geoclue_emit_status_changed(GC_IFACE_GEOCLUE(provider), GEOCLUE_STATUS_AVAILABLE);
+	}
+
+	return TRUE;
+}
+
 /* Called when the message is finished in its response */
 static void
 message_finished (SoupMessage * message, gpointer user_data)
@@ -493,69 +664,12 @@ message_finished (SoupMessage * message, gpointer user_data)
 		return;
 	}
 
-	provider->priv->markup_data.position_changed = FALSE;
-	provider->priv->markup_data.address_changed = FALSE;
-	provider->priv->markup_data.valid_header = FALSE;
-
-	if (g_markup_parse_context_parse(provider->priv->context, message->response_body->data, message->response_body->length, NULL) == FALSE) {
-		g_warning("Unable to parse the data from the geoip provider");
-
-		g_clear_object(&provider->priv->context);
-		provider->priv->context = g_markup_parse_context_new(&parser_struct, 0, &provider->priv->markup_data, NULL);
-
+	if (parse_xml_data(provider, message->response_body->data, message->response_body->length)) {
+		/* If we got something valid, reset our progressive backoff */
+		provider->priv->requeue_time = REQUEUE_MIN_SEC;
+	} else {
+		/* If we got junk, try again later */
 		queue_message(provider);
-		return;
-	}
-
-	if (!provider->priv->markup_data.valid_header) {
-		g_warning("We didn't get valid XML, that's not cool.");
-		queue_message(provider);
-		return;
-	}
-
-	provider->priv->requeue_time = REQUEUE_MIN_SEC;
-
-	gboolean emit = FALSE;
-	if (provider->priv->gotdata == FALSE) {
-		emit = TRUE;
-	}
-	provider->priv->gotdata = TRUE;
-
-	if (provider->priv->markup_data.position_changed) {
-		g_debug("New position:");
-		g_debug("Latitude:  %f", provider->priv->latitude);
-		g_debug("Longitude: %f", provider->priv->longitude);
-
-		GeoclueAccuracy * accuracy = geoclue_accuracy_new (GEOCLUE_ACCURACY_LEVEL_LOCALITY, 0.0, 0.0);
-		gc_iface_position_emit_position_changed(GC_IFACE_POSITION(provider), 
-		                                        time(NULL),
-		                                        GEOCLUE_POSITION_FIELDS_LATITUDE | GEOCLUE_POSITION_FIELDS_LONGITUDE,
-		                                        provider->priv->latitude,
-		                                        provider->priv->longitude,
-		                                        0, /* Altitude */
-		                                        accuracy);
-
-		geoclue_accuracy_free(accuracy);
-	}
-
-	if (provider->priv->markup_data.address_changed) {
-		GeoclueAccuracy * accuracy = geoclue_accuracy_new (GEOCLUE_ACCURACY_LEVEL_LOCALITY, 0.0, 0.0);
-		GHashTable * address = geoclue_address_details_new();
-
-		address_details_build(provider, address);
-
-		gc_iface_address_emit_address_changed(GC_IFACE_ADDRESS(provider), 
-		                                        time(NULL),
-		                                        address,
-		                                        accuracy);
-
-		g_hash_table_destroy(address);
-		geoclue_accuracy_free(accuracy);
-	}
-
-	if (emit) {
-		g_debug("Emitting Available");
-		gc_iface_geoclue_emit_status_changed(GC_IFACE_GEOCLUE(provider), GEOCLUE_STATUS_AVAILABLE);
 	}
 
 	return;
@@ -829,6 +943,11 @@ markup_text (GMarkupParseContext * context, const gchar * text, gsize text_len,
 		return;
 	}
 
+	if (text_len == 0) {
+		/* Null strings aren't interesting to us */
+		return;
+	}
+
 	/* Decode the enum into an offset */
 	switch (markup_data->field) {
 	case MARKUP_FIELD_COUNTRY_CODE:
@@ -866,6 +985,11 @@ markup_text (GMarkupParseContext * context, const gchar * text, gsize text_len,
 		newstr[text_len] = '\0';
 	}
 
+	if (g_strcmp0(newstr, "None") == 0) {
+		g_free(newstr);
+		return;
+	}
+
 	/* If we've got a string field, let's deal with that */
 	if (field != NULL && newstr != NULL) {
 		/* We've got new data? */
@@ -897,13 +1021,30 @@ markup_text (GMarkupParseContext * context, const gchar * text, gsize text_len,
 	return;
 }
 
+static GOptionEntry general_options[] = {
+	{"test-data",  0,  0,  G_OPTION_ARG_FILENAME,  &test_data,  "Path to test data to use instead of getting it from the network", "XML Data File"},
+	{NULL}
+};
 
 /* Function to start everything */
 int
-main (int argv, char * argc[])
+main (int argc, char * argv[])
 {
+	GError * error = NULL;
+	GOptionContext * context;
+
 	g_type_init();
 
+	context = g_option_context_new("- Get the location of the machine using GeoIP and report it through GeoClue");
+
+	g_option_context_add_main_entries(context, general_options, "dbus-runner");
+
+	if (!g_option_context_parse(context, &argc, &argv, &error)) {
+		g_print("option parsing failed: %s\n", error->message);
+		g_error_free(error);
+		return 1;
+	}
+
 	UbuntuGeoipProvider * provider = g_object_new(UBUNTU_GEOIP_PROVIDER_TYPE, NULL);
 	g_return_val_if_fail(provider != NULL, 1);
 
